import os
import sys
from copy import deepcopy
from .style import Raw


def outlet(handler):
    method = staticmethod(handler)
    return type(handler.__name__, (Outlet,), {"_handler": method})


class Outlet(object):

    def __init__(self, style=Raw(), *args, **kwargs):
        self._style = style
        self._args = args
        self._kwargs = kwargs

    def __call__(self, log):
        if self._style is not None:
            log = self._style(log)
        if log is not None:
            self._handler(log, *self._args, **self._kwargs)

    def _handler(self, *args, **kwargs):
        raise NotImplementedError()


@outlet
def File(log, filepath):
    filepath = os.path.realpath(os.path.expanduser(filepath))
    with open(filepath, "a+") as f:
        f.write(log)


@outlet
def Print(log):
    sys.stdout.write(log)



import sys
import json
import datetime
import inspect


class Style(object):

    default = None

    def __call__(self, log):
        method = getattr(self, log["metadata"]["status"], self.default)
        if method is not None:
            lines = method(log)
            if lines and isinstance(lines, str):
                lines = [lines]
            else:
                lines = list(lines)
            if lines:
                return "\n".join(lines) + "\n"


class Raw(Style):

    def default(self, log):
        try:
            return json.dumps(log)
        except:
            content = {k: str(v) for k, v in log["content"].items()}
            return json.dumps(dict(log, content=content))


class Log(Style):

    def working(self, log):
        if "reason" in log["content"]:
            metadata = log["metadata"]
            status = metadata["status"]
            timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
            info = "mem: %{:.6}, cpu: %{:.6}".format(log["mem"], log["cpu"])
            yield "{time} [{status}] {reason} - {info}".format(
                time=timestamp, status=status,
                reason=log["content"]["reason"], info=info)

    def success(self, log):
        if "reason" in log["content"]:
            metadata = log["metadata"]
            status = metadata["status"]
            duration = metadata["stop"] - metadata["start"]
            timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
            info = "mem: %{:.6}, cpu: %{:.6}".format(log["mem"], log["cpu"])
            yield "{time} [{status}] {reason} after {duration:.3f} seconds - {info}".format(
                time=timestamp, status=status, duration=duration,
                reason=log["content"]["reason"], info=info)

    def failure(self, log):
        if "reason" in log["content"]:
            metadata = log["metadata"]
            status = metadata["status"]
            duration = metadata["stop"] - metadata["start"]
            timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
            info = "mem: %{:.6}, cpu: %{:.6}".format(log["mem"], log["cpu"])
            yield "{time} [{status}] {reason} after {duration:.3f} seconds - {info}".format(
                time=timestamp, status=status, duration=duration,
                reason=log["content"]["reason"], info=info)


class Tree(Style):

    def started(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * metadata["depth"]
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "|-- {status}: {title}".format(**metadata)
        yield indent + "|   @ {0}".format(timestamp)
        for k, v in content.items():
            yield indent + "|   | {0}: {1}".format(k, v)

    def working(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "|-- {status}: {title}".format(**metadata)
        yield indent + "|   @ {0}".format(timestamp)
        for k, v in content.items():
            yield indent + "|   | {0}: {1}".format(k, v)

    def success(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "`-- {status}: {title}".format(**metadata)
        yield indent + "    @ {0}".format(timestamp)
        diff = metadata["stop"] - metadata["start"]
        content["duration"] = "{:.3f}".format(diff) + " seconds"
        for k, v in content.items():
            yield indent + "    | {0}: {1}".format(k, v)

    def failure(self, log):
        content, metadata = log["content"], log["metadata"]
        indent = "|   " * (metadata["depth"] + 1)
        timestamp = datetime.datetime.fromtimestamp(log["timestamp"])
        yield indent + "`-- {status}: {title}".format(**metadata)
        yield indent + "    @ {0}".format(timestamp)
        diff = metadata["stop"] - metadata["start"]
        content["duration"] = "{:.3f}".format(diff) + " seconds"
        if content["reason"] in content:
            del content["reason"]
        for k, v in content.items():
            yield indent + "    | {0}: {1}".format(k, v)



import sys
import os
import json
import time
import types
import psutil
import inspect
import asyncio
import threading
import functools
from uuid import uuid1
from weakref import WeakKeyDictionary

from .utils import CallMap, infer_title
from .dist import DistributorThread

if sys.version_info >= (3, 5):
    from ._shelf_py35 import _book_compat
else:
    from ._shelf_py27 import _book_compat


class book(dict, _book_compat):

    _shelves = WeakKeyDictionary()
    _distributor = DistributorThread()

    @classmethod
    def distribute(cls, *outlets):
        for o in outlets:
            cls._distributor.add_outlet(o)

    def __init__(self, title=None, parent=None, **content):
        parent = parent or self.current().get("tag")
        depth = int(parent.split("-")[1]) + 1 if parent else 0
        title = title or "%s - task" % self.current().get("title")
        super().__init__(tag=uuid1().hex + "-%s" % depth, depth=depth,
            start=time.time(), status="started", parent=parent, title=title)
        self._write(content)
        self._conclusion = {}

    def __enter__(self):
        self.shelf().append(self)
        self.update(status="working")
        return self

    def __exit__(self, *exc):
        self["stop"] = time.time()
        if exc[0] is not None:
            etype = exc[0].__name__
            self.update(status="failure")
            self._conclusion[etype] = str(exc[1])
        else:
            self.update(status="success")
        self._write(self._conclusion)
        self.shelf().pop()
        return False

    def resume(self, tag):
        return self.bind(parent=tag)

    @classmethod
    def write(cls, *args, **kwargs):
        content = dict(*args, **kwargs)
        cls.current()._write(content)

    @classmethod
    def close(cls, *args, **kwargs):
        content = dict(*args, **kwargs)
        cls.current()._conclusion.update(content)

    def _write(self, content):
        process = psutil.Process(os.getpid())
        log = {
            "metadata": self.copy(),
            "content": content,
            "timestamp": time.time(),
            "mem": process.memory_percent(),
            "cpu": process.cpu_percent(),
        }
        self._distributor(log)

    @classmethod
    def current(cls):
        return cls.shelf()[-1]



import inspect
import functools
import threading

from .utils import CallMap, infer_title


class _book_compat(object):

    @classmethod
    def shelf(cls):
        task = threading.current_thread()
        if task not in cls._shelves:
            cls._shelves[task] = [{"title": "__main__"}]
        return cls._shelves[task]

    @classmethod
    def mark(cls, title, *binding, **content):
        def setup(function, title=title):
            if title is None:
                title = infer_title(function)
            cm = CallMap(function)
            @functools.wraps(function)
            def author(*args, **kwargs):
                book = kwargs.pop("__book__", None)
                intro = dict(content, **cm.map(args, kwargs))
                with book or cls(title, *binding, **intro):
                    result = function(*args, **kwargs)
                    cls.close({"returned": result})
                    return result
            return author
        if callable(title):
            return setup(title, None)
        else:
            return setup



import json
from .via import Print
from .style import Tree


class Stream(object):

    def __init__(self, outlet=Print(Tree())):
        self._logs = {} # series of log queues per tag
        self._tree = {} # the tag paths to each log
        self._sending = None
        self._pending = []
        self._outlet = outlet

    def __call__(self, log):
        if isinstance(log, str):
            log = json.loads(log)
        status = log["metadata"].get("status")
        if status == "started":
            self._start(log)
        elif status == "working":
            self._continue(log)
        elif status in ("success", "failure"):
            self._close(log)
        else:
            raise ValueError("Invalid status %r" % status)

    def _start(self, log):
        metadata = log["metadata"]
        tag = metadata["tag"]
        parent = metadata["parent"]
        self._logs[tag] = [log]
        tree = self._tree.setdefault(parent, [])
        self._tree[tag] = tree + [tag]

    def _continue(self, log):
        tag = log["metadata"]["tag"]
        self._logs[tag].append(log)

    def _close(self, log):
        tag = log["metadata"]["tag"]
        if self._sending is None:
            self._send(log)
        elif self._sending == self._tree[tag][0]:
            self._send(log)
        else:
            self._pending.append(log)

    def _send(self, log):
        tag = log["metadata"]["tag"]
        for t in self._tree[tag]:
            queue = self._logs[t]
            for _ in range(len(queue)):
                self._outlet(queue.pop(0))
        self._outlet(log)
        if log["metadata"]["parent"] is None:
            self._sending = None
            if self._pending:
                self._send(self._pending.pop(0))
        else:
            self._sending = self._tree[tag][0]
        del self._logs[tag]
        del self._tree[tag]



import os
import sys
import subprocess
from os import path
from glob import glob
from distutils.core import setup
from setuptools import find_packages

#-----------------------------------------------------------------------------
# Project
#-----------------------------------------------------------------------------

url = "https://github.com/rmorshea/jotting"
project = "jotting"
version = "0.1.1"
author = "Ryan Morshead"
email = "ryan.morshead@gmail.com"
summary = "causally related logging messages"

description = """
Jotting
-------

Logs that explain when, where, and why things happen.

``jotting`` is a log system for Python 3 that can be used to record the causal
history of an asynchronous or distributed system. These histories are composed
of actions which, once "started", will begin "working", potentially spawn other
actions, and eventually end as a "success" or "failure". In the end you're left
with a breadcrumb trail of information that you can use to squash bugs with
minimal boilerplate.

Jotting was heavily inspired by ``eliot`` (https://eliot.readthedocs.io/).
"""

#-----------------------------------------------------------------------------
# Packages
#-----------------------------------------------------------------------------

packages = find_packages(project)

#-----------------------------------------------------------------------------
# Base Paths
#-----------------------------------------------------------------------------

here = path.abspath(path.dirname(__file__))
root = path.join(here, project)

#-----------------------------------------------------------------------------
# Finalize Parameters
#-----------------------------------------------------------------------------

if sys.version_info >= (3, 5):
    requires = ["psutil"]
else:
    requires = ["psutil", "funcsigs"]

parameters = dict(
    url=url,
    name=project,
    version=version,
    author=author,
    author_email=email,
    description=summary,
    long_description=description,
    packages=find_packages(),
    python_requires = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    install_requries=requires,
    classifiers = [
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)

#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

if __name__ == "__main__":
    setup(**parameters)
